@@ -4,6 +4,7 @@
 #include <Rcpp.h>
 #include <queue>
 #include <boost/function.hpp>
+#include <boost/shared_ptr.hpp>
 #include "timestamp.h"
 #include "optional.h"
 #include "threadutils.h"
@@ -33,10 +34,24 @@ private:
   Task func;
 };
 
+typedef boost::shared_ptr<Callback> Callback_sp;
+
+template <typename T>
+struct pointer_greater_than {
+  const bool operator()(const T a, const T b) const {
+    return *a > *b;
+  }
+};
+
+
 // Stores R function callbacks, ordered by timestamp.
 class CallbackRegistry {
 private:
-  std::priority_queue<Callback,std::vector<Callback>,std::greater<Callback> > queue;
+  // This is a priority queue of shared pointers to Callback objects. The
+  // reason it is not a priority_queue<Callback> is because that can cause
+  // objects to be copied on the wrong thread, and even trigger an R GC event
+  // on the wrong thread. https://github.com/r-lib/later/issues/39
+  std::priority_queue<Callback_sp, std::vector<Callback_sp>, pointer_greater_than<Callback_sp> > queue;
   mutable Mutex mutex;
   mutable ConditionVariable condvar;
 
@@ -62,7 +77,7 @@ public:
   bool due(const Timestamp& time = Timestamp()) const;
   
   // Pop and return an ordered list of functions to execute now.
-  std::vector<Callback> take(size_t max = -1, const Timestamp& time = Timestamp());
+  std::vector<Callback_sp> take(size_t max = -1, const Timestamp& time = Timestamp());
   
   bool wait(double timeoutSecs) const;
 };
